'''
Created on Mar 27, 2017

@author: nicolas
'''

from __future__ import annotations

import re

from typing import Optional, Any

CONJUGATION_FORMS = {
    re.compile(r"^(to be)"): ("is", "is not", "to not be"),
    re.compile(r"^(to have)"): ("has", "has not", "to have no"),
    re.compile(r"^(to match)"): ("matches", "does not match", "to not match"),
    re.compile(r"^(can)"): ("can", "cannot", "to cannot")
}


class MatcherDescriptionTransformer:
    """
    This class is used as a callable and passed to :py:meth:`Matcher.build_description`
    to transform the leading verb in description according to the transformer settings.
    """
    def __init__(self, conjugate=False, negative=False):
        #: indicate whether or not the verb in the description will be conjugated
        self.conjugate = conjugate
        #: indicate whether or not the description will be turned into the negative form
        self.negative = negative

    def __call__(self, description: str) -> str:
        """
        Transform the description according transformer settings.
        """
        ###
        # No transformation to do
        ###
        if not self.conjugate and not self.negative:
            return description

        ###
        # Transformation of irregular verb
        ###
        for pattern, forms in CONJUGATION_FORMS.items():
            conjugated, conjugated_negative, infinitive_negative = forms
            if self.conjugate and self.negative:
                substitution = conjugated_negative
            elif self.conjugate:
                substitution = conjugated
            else:  # self.negative
                substitution = infinitive_negative

            if pattern.match(description):
                return pattern.sub(substitution, description)

        ###
        # Transformation of regular verb
        ###
        pattern = re.compile(r"^to (\w+)")
        m = pattern.match(description)
        if m:
            verb = m.group(1)
            if self.conjugate and self.negative:
                substitution = "does not " + verb
            elif self.conjugate:
                substitution = verb + "s"
            elif self.negative:
                substitution = "to not " + verb
            else:
                substitution = "not " + verb

            return pattern.sub(substitution, description)

        ###
        # No verb detected
        ###
        return description


# In 1.1.0, MatchDescriptionTransformer has been renamed into MatcherDescriptionTransformer
# keep a MatchDescriptionTransformer alias for backward-compatibility
MatchDescriptionTransformer = MatcherDescriptionTransformer


class MatchResult:
    def __init__(self, is_successful: bool, description: str = None) -> None:
        #: whether or not the match did succeed
        self.is_successful: bool = is_successful
        #: optional description
        self.description: str = description

    @classmethod
    def success(cls, description: str = None) -> MatchResult:
        """
        Shortcut used to create a "successful" MatchResult.
        """
        return cls(True, description)

    @classmethod
    def failure(cls, description: str = None) -> MatchResult:
        """
        Shortcut used to create a "failed" MatchResult.
        """
        return cls(False, description)

    def __bool__(self):
        """
        Returns whether or not the match is successful.
        """
        return self.is_successful


class Matcher:
    def build_description(self, transformation: MatcherDescriptionTransformer) -> str:
        """
        Build a description for the matcher given the instance of :py:class:`MatcherDescriptionTransformer`
        passed as argument.
        """
        raise NotImplementedError()

    def build_short_description(self, transformation):
        return self.build_description(transformation)

    def matches(self, actual: Any) -> MatchResult:
        """
        Test if the passed argument matches.

        :param actual: the actual value to match
        :return: an instance of :py:class:`MatchResult`
        """
        raise NotImplementedError()

    def override_description(self, description: str) -> Matcher:
        """
        Override the matcher description.

        :param description: the new description
        :return: a new matcher wrapping the actual matcher
        """
        return MatcherWrapper(self, description=description)

    def hide_result_details(self) -> Matcher:
        """
        Hide the matching operation result details.

        :return: a new matcher wrapping the actual matcher
        """
        return MatcherWrapper(self, result_details=None)


class MatcherWrapper(Matcher):
    def __init__(self, matcher, description: str = NotImplemented, result_details: Optional[str] = NotImplemented) -> None:
        self.matcher = matcher
        self.description = description
        self.result_details = result_details

    def build_description(self, transformation):
        if self.description is NotImplemented:
            return self.matcher.build_description(transformation)
        else:
            return transformation(self.description)

    def matches(self, actual):
        result = self.matcher.matches(actual)
        if self.result_details is not NotImplemented:
            result.description = self.result_details
        return result



import inspect

from .matchers import *
from .operations import check_that, require_that, assert_that, check_that_in, require_that_in, assert_that_in

DISPLAY_DETAILS_WHEN_EQUAL = True


def _is_symbol_exportable(sym_name):
    return all((
        sym_name != "DISPLAY_DETAILS_WHEN_EQUAL",
        not sym_name.startswith("_"),
        not inspect.ismodule(globals()[sym_name])
    ))


__all__ = list(filter(_is_symbol_exportable, dir()))



from lemoncheesecake.project import load_project, PreparedProject
from lemoncheesecake.cli.command import Command
from lemoncheesecake.cli.utils import add_project_cli_arg


class CheckCommand(Command):
    def get_name(self):
        return "check"

    def get_description(self):
        return "Check if the project is valid (project module, suites, fixtures, metadata policy)"

    def add_cli_args(self, cli_parser):
        add_project_cli_arg(cli_parser)

    def run_cmd(self, cli_args):
        project = load_project(cli_args.project)
        PreparedProject.create(project)

        print("Everything is ok")



'''
Created on Jan 1, 2017

@author: nicolas
'''

import os
import platform
import time
import glob
import re
import shutil

DEFAULT_REPORT_DIR_NAME = "report"


def archive_dirname_datetime(ts, archives_dir):
    return time.strftime("report-%Y%m%d-%H%M%S", time.localtime(ts))


# TODO: create two different functions
def report_dir_with_archiving(top_dir, archive_dirname_callback):
    archives_dir = os.path.join(top_dir, "reports")

    if platform.system() == "Windows":
        report_dir = os.path.join(top_dir, DEFAULT_REPORT_DIR_NAME)
        if os.path.exists(report_dir):
            if not os.path.exists(archives_dir):
                os.mkdir(archives_dir)
            os.rename(
                report_dir, os.path.join(archives_dir, archive_dirname_callback(os.path.getctime(report_dir), archives_dir))
            )
        os.mkdir(report_dir)

    else:
        if not os.path.exists(archives_dir):
            os.mkdir(archives_dir)

        report_dirname = archive_dirname_callback(time.time(), archives_dir)

        report_dir = os.path.join(archives_dir, report_dirname)
        os.mkdir(report_dir)

        symlink_path = os.path.join(os.path.dirname(report_dir), "..", DEFAULT_REPORT_DIR_NAME)
        if os.path.lexists(symlink_path):
            os.unlink(symlink_path)
        os.symlink(report_dir, symlink_path)

    return report_dir


def _list_directories_for_rotation(top_dir, dir_prefix):
    dirs = {}
    for dirname in glob.glob(os.path.join(top_dir, "%s*" % dir_prefix)):
        m = re.compile(r"^%s(\d+)$" % dir_prefix).search(os.path.basename(dirname))
        if m is not None:
            dirs[int(m.group(1))] = dirname
    return dirs


def _remove_obsolete_directories(directories, limit):
    if limit is None or len(list(filter(lambda num: num <= limit, directories.keys()))) < limit:
        return directories

    new_dirs = {}
    for dir_num, dir_path in directories.items():
        if dir_num >= limit:
            # directory could be a link if is has been previously created by report_dir_with_archiving:
            if os.path.islink(dir_path):
                os.remove(dir_path)
            else:
                shutil.rmtree(dir_path)
        else:
            new_dirs[dir_num] = dir_path
    return new_dirs


def _rotate_directory(num, dirname):
    next_num = num + 1
    next_dirname = re.sub(r"\d+$", str(next_num), dirname)
    if os.path.exists(next_dirname):
        _rotate_directory(next_num, next_dirname)
    os.rename(dirname, next_dirname)


def _rotate_directories(directories):
    if 1 in directories:
        _rotate_directory(1, directories[1])


def create_report_dir_with_rotation(top_dir, archiving_limit=20):
    report_dir = os.path.join(top_dir, DEFAULT_REPORT_DIR_NAME)
    if os.path.exists(report_dir):
        archiving_dir = os.path.join(top_dir, "reports")
        if not os.path.exists(archiving_dir):
            os.mkdir(archiving_dir)
        _rotate_directories(
            _remove_obsolete_directories(
                _list_directories_for_rotation(archiving_dir, "report-"), limit=archiving_limit
            )
        )
        os.rename(report_dir, os.path.join(archiving_dir, "report-1"))

    os.mkdir(report_dir)

    return report_dir



from .value import equal_to, not_equal_to, greater_than, greater_than_or_equal_to, \
    less_than, less_than_or_equal_to, is_between, is_none, is_not_none, has_length, is_true, is_false
from .string import starts_with, ends_with, contains_string, match_pattern, is_text, is_json
from .list_ import has_item, has_items, has_only_items, has_all_items, is_in
from .dict_ import has_entry
from .types_ import is_integer, is_float, is_bool, is_str, is_dict, is_list
from .composites import all_of, any_of, anything, something, existing, present, is_, not_, is_not



import os
import sys
import glob
import fnmatch
import re
import importlib.util

from lemoncheesecake.exceptions import serialize_current_exception, ModuleImportError


def strip_py_ext(filename):
    return re.sub(r"\.py$", "", filename)


def get_py_files_from_dir(dir):
    return \
        sorted(
            filter(
                lambda f: not os.path.basename(f).startswith("__"), glob.glob(os.path.join(dir, "*.py"))
            )
        )


def get_matching_files(patterns, excluding=[]):
    if type(patterns) not in (list, tuple):
        patterns = [patterns]
    if type(excluding) not in (list, tuple):
        excluding = [excluding]
    files = []
    for pattern in patterns:
        files.extend(glob.glob(pattern))
    if excluding:
        tmp = files[:]  # iterate on copy to be able to alter files
        for file in tmp:
            for excluded in excluding:
                if fnmatch.fnmatch(file, excluded):
                    files.remove(file)
                    break
    return sorted(files)


def import_module(path):
    try:
        # NB: path is used as module name to avoid possible conflicting with Python modules using
        # the same name
        spec = importlib.util.spec_from_file_location(path, path)
        module = importlib.util.module_from_spec(spec)
        sys.modules[path] = module
        spec.loader.exec_module(module)
    except Exception:
        raise ModuleImportError(
            "Error while importing file '%s': %s" % (path, serialize_current_exception(show_stacktrace=True))
        )

    return module



import os
import os.path as osp
import sys
import traceback
import mimetypes

try:
    import reportportal_client
    REPORT_PORTAL_CLIENT_IS_AVAILABLE = True
except ImportError:
    REPORT_PORTAL_CLIENT_IS_AVAILABLE = False

from lemoncheesecake.reporting.backend import ReportingBackend, ReportingSession, ReportingSessionBuilderMixin
from lemoncheesecake.exceptions import UserError
from lemoncheesecake.events import SyncEventManager
from lemoncheesecake.reporting.replay import replay_report_events


def make_time(t):
    return str(int(t * 1000))


def make_tags_from_test_tree_node(node):
    return node.tags + \
           ["%s_%s" % (name, value) for name, value in node.properties.items()] + \
           [link[1] or link[0] for link in node.links]


class ReportPortalReportingSession(ReportingSession):
    def __init__(self, url, auth_token, project, launch_name, launch_description, report_dir, report):
        self.service = reportportal_client.ReportPortalServiceAsync(
            endpoint=url, project=project, token=auth_token, error_handler=self._handle_rp_error
        )
        self.launch_name = launch_name
        self.launch_description = launch_description
        self.report_dir = report_dir
        self.report = report
        self._rp_exc_info = None

    def _handle_rp_error(self, exc_info):
        self._rp_exc_info = exc_info
        return False  # stop on error

    def _has_rp_error(self):
        return self._rp_exc_info is not None

    def _show_rp_error(self):
        print(
            "Got the following exception using ReportPortal, "
            "test results have not been properly synced:",
            file=sys.stderr
        )
        traceback.print_exception(*self._rp_exc_info, file=sys.stderr)

    def _end_current_test_item(self, end_time, status):
        self.service.finish_test_item(end_time=make_time(end_time), status=status)

    def _start_test_item(self, item_type, start_time, name, description, wrapped=False):
        if wrapped:
            self.service.start_test_item(
                item_type="SUITE", start_time=make_time(start_time),
                name=name, description=description
            )
        self.service.start_test_item(
            item_type=item_type, start_time=make_time(start_time),
            name=name, description=description
        )

    def _end_test_item(self, end_time, is_successful, wrapped=False):
        status = "passed" if is_successful else "failed"
        if wrapped:
            self._end_current_test_item(end_time, status=status)
        self._end_current_test_item(end_time, status=status)

    def on_test_session_start(self, event):
        if self._has_rp_error():
            return

        self.service.start_launch(
            name=self.launch_name, description=self.launch_description, start_time=make_time(event.time)
        )

    def on_test_session_end(self, event):
        if self._has_rp_error():
            self._show_rp_error()
        else:
            self.service.finish_launch(end_time=make_time(event.time))
            self.service.terminate()

            if self._has_rp_error():
                self._show_rp_error()

    def on_test_session_setup_start(self, event):
        if self._has_rp_error():
            return

        self._start_test_item(
            item_type="BEFORE_CLASS", start_time=event.time,
            name="session_setup", description="Test Session Setup",
            wrapped=True
        )

    def on_test_session_setup_end(self, event):
        if self._has_rp_error():
            return

        self._end_test_item(
            event.time,
            not self.report.test_session_setup or self.report.test_session_setup.is_successful(),
            wrapped=True
        )

    def on_test_session_teardown_start(self, event):
        if self._has_rp_error():
            return

        self._start_test_item(
            item_type="AFTER_CLASS", start_time=event.time,
            name="session_teardown", description="Test Session Teardown",
            wrapped=True
        )

    def on_test_session_teardown_end(self, event):
        if self._has_rp_error():
            return

        self._end_test_item(
            event.time,
            not self.report.test_session_teardown or self.report.test_session_teardown.is_successful(),
            wrapped=True
        )

    def on_suite_start(self, event):
        if self._has_rp_error():
            return

        suite = event.suite
        self.service.start_test_item(
            item_type="SUITE", start_time=make_time(event.time),
            name=suite.name, description=suite.description,
            tags=make_tags_from_test_tree_node(suite)
        )

    def on_suite_end(self, event):
        if self._has_rp_error():
            return

        self._end_current_test_item(event.time, status="passed")

    def on_suite_setup_start(self, event):
        if self._has_rp_error():
            return

        self._start_test_item(
            item_type="BEFORE_CLASS", start_time=event.time,
            name="suite_setup", description="Suite Setup",
            wrapped=len(event.suite.get_suites()) > 0
        )

    def on_suite_setup_end(self, event):
        if self._has_rp_error():
            return

        suite_data = self.report.get_suite(event.suite)

        self._end_test_item(
            event.time,
            not suite_data.suite_setup or suite_data.suite_setup.is_successful(),
            wrapped=len(event.suite.get_suites()) > 0
        )

    def on_suite_teardown_start(self, event):
        if self._has_rp_error():
            return

        self._start_test_item(
            item_type="AFTER_CLASS", start_time=event.time,
            name="suite_teardown", description="Suite Teardown",
            wrapped=len(event.suite.get_suites()) > 0
        )

    def on_suite_teardown_end(self, event):
        if self._has_rp_error():
            return

        suite_data = self.report.get_suite(event.suite)

        self._end_test_item(
            event.time,
            not suite_data.suite_teardown or suite_data.suite_teardown.is_successful(),
            wrapped=len(event.suite.get_suites()) > 0
        )

    def on_test_start(self, event):
        if self._has_rp_error():
            return

        test = event.test
        self.service.start_test_item(
            item_type="TEST", start_time=make_time(event.time),
            name=test.name, description=test.description,
            tags=make_tags_from_test_tree_node(test)
        )

    def on_test_end(self, event):
        if self._has_rp_error():
            return

        test_data = self.report.get_test(event.test)
        self._end_current_test_item(event.time, test_data.status)

    def _bypass_test(self, test, status, time):
        if self._has_rp_error():
            return

        self.service.start_test_item(
            item_type="TEST", start_time=make_time(time),
            name=test.name, description=test.description, tags=test.tags,
        )
        self._end_current_test_item(time, status=status)

    def on_test_skipped(self, event):
        if self._has_rp_error():
            return

        self._bypass_test(event.test, "skipped", event.time)

    def on_disabled_test(self, event):
        # do not log disabled test, moreover it seems that there is not corresponding status in ReportPortal
        pass

    def on_step_start(self, event):
        if self._has_rp_error():
            return

        self.service.log(make_time(event.time), "--- STEP: %s ---" % event.step_description, "INFO")

    def on_log(self, event):
        if self._has_rp_error():
            return

        self.service.log(make_time(event.time), event.log_message, event.log_level.upper())

    def on_check(self, event):
        if self._has_rp_error():
            return

        message = "%s => %s" % (event.check_description, "OK" if event.check_is_successful else "NOT OK")
        if event.check_details is not None:
            message += "\nDetails: %s" % event.check_details
        self.service.log(make_time(event.time), message, "INFO" if event.check_is_successful else "ERROR")

    def on_log_attachment(self, event):
        if self._has_rp_error():
            return

        abspath = os.path.join(self.report_dir, event.attachment_path)
        with open(abspath, "rb") as fh:
            self.service.log(make_time(event.time), event.attachment_description, "INFO", attachment={
                "name": osp.basename(event.attachment_path),
                "data": fh.read(),
                "mime": mimetypes.guess_type(abspath)[0] or "application/octet-stream"
            })

    def on_log_url(self, event):
        if self._has_rp_error():
            return

        if event.url_description and event.url_description != event.url:
            message = "%s: %s" % (event.url_description, event.url)
        else:
            message = event.url
        self.service.log(make_time(event.time), message, "INFO")


class ReportPortalReportingSessionParallelized(ReportingSession):
    def __init__(self, *args):
        self._session = ReportPortalReportingSession(*args)

    def on_test_session_end(self, _):
        event_manager = SyncEventManager.load()
        event_manager.add_listener(self._session)
        replay_report_events(self._session.report, event_manager)


class ReportPortalBackend(ReportingBackend, ReportingSessionBuilderMixin):
    def get_name(self):
        return "reportportal"

    def is_available(self):
        return REPORT_PORTAL_CLIENT_IS_AVAILABLE

    def create_reporting_session(self, report_dir, report, parallel, _):
        try:
            url = os.environ["LCC_RP_URL"]
            auth_token = os.environ["LCC_RP_AUTH_TOKEN"]
            project = os.environ["LCC_RP_PROJECT"]
            launch_name = os.environ.get("LCC_RP_LAUNCH_NAME", "Test Run")
            launch_description = os.environ.get("LCC_RP_LAUNCH_DESCRIPTION", None)
        except KeyError as excp:
            raise UserError("ReportPortal reporting backend, cannot get environment variable %s" % excp)

        # the ReportPortal REST API allows working on multiple test item at a time,
        # unfortunately reportportal_client module does not support it because test item ids are not exposed and
        # the API make it impossible to work on multiple test item in parallel.
        # When lemoncheesecake tests are run in parallel, we must then wait the end of the test session to
        # replay the events from the report as the tests would have been run sequentially.
        if parallel:
            return ReportPortalReportingSessionParallelized(
                url, auth_token, project, launch_name, launch_description, report_dir, report
            )
        else:
            return ReportPortalReportingSession(
                url, auth_token, project, launch_name, launch_description, report_dir, report
            )



'''
Created on May 2, 2017

@author: nicolas
'''

from typing import Any, Collection

from lemoncheesecake.helpers.text import jsonify
from lemoncheesecake.matching.matcher import Matcher, MatchResult, MatcherDescriptionTransformer
from lemoncheesecake.matching.matchers.composites import is_


def _jsonify_items(items):
    return ", ".join(map(jsonify, items))


class HasItemMatchResult(MatchResult):
    def __init__(self, is_successful, description, index, item):
        MatchResult.__init__(self, is_successful, description)
        self.index: int = index
        self.item: Any = item

    @classmethod
    def found(cls, index, item):
        return cls(True, "found matching item at index %d" % index, index, item)

    @classmethod
    def not_found(cls):
        return cls(False, "no matching item", -1, None)


class HasItem(Matcher):
    def __init__(self, expected):
        self.expected = expected

    def build_description(self, transformation):
        return transformation(
            "to have an item whose value %s" % self.expected.build_description(MatcherDescriptionTransformer(conjugate=True))
        )

    def build_short_description(self, transformation):
        return transformation(
            "to have an item whose value %s" % self.expected.build_short_description(MatcherDescriptionTransformer(conjugate=True))
        )

    def matches(self, actual):
        for index, item in enumerate(actual):
            result = self.expected.matches(item)
            if result:
                return HasItemMatchResult.found(index, item)
        return HasItemMatchResult.not_found()


def has_item(expected: Any) -> Matcher:
    """Test if the sequence has item matching expected"""
    return HasItem(is_(expected))


class HasItems(Matcher):
    def __init__(self, expected):
        self.expected = expected

    def build_description(self, transformation):
        return transformation("to have items %s" % _jsonify_items(self.expected))

    def matches(self, actual):
        missing = []
        for expected in self.expected:
            if expected not in actual:
                missing.append(expected)

        if missing:
            return MatchResult.failure("Missing items: %s" % _jsonify_items(missing))
        else:
            return MatchResult.success("got %s" % jsonify(actual))


def has_items(values: Collection) -> Matcher:
    """Test if the sequence contains at least the given values"""
    return HasItems(values)


class HasOnlyItems(Matcher):
    def __init__(self, expected):
        self.expected = expected

    def build_description(self, transformation):
        return transformation("to have only items %s" % _jsonify_items(self.expected))

    def matches(self, actual):
        expected = list(self.expected)
        extra = []
        for value in actual:
            if value in expected:
                expected.remove(value)
            else:
                extra.append(value)

        if len(expected) == 0 and len(extra) == 0:
            return MatchResult.success("got %s" % jsonify(actual))
        else:
            details = []
            if len(expected) > 0:
                details.append("Missing items: %s" % _jsonify_items(expected))
            if len(extra) > 0:
                details.append("Extra items: %s" % _jsonify_items(extra))
            return MatchResult.failure("; ".join(details))


def has_only_items(expected: Collection) -> Matcher:
    """Test if the sequence only contains the given values"""
    return HasOnlyItems(expected)


class HasAllItems(Matcher):
    def __init__(self, expected):
        self.expected = expected

    def build_description(self, transformation):
        return transformation(
            "to have all items whose value %s" %
            self.expected.build_description(MatcherDescriptionTransformer(conjugate=True))
        )

    def build_short_description(self, transformation):
        return transformation(
            "to have all items whose value %s" %
            self.expected.build_short_description(MatcherDescriptionTransformer(conjugate=True))
        )

    def matches(self, actual):
        failures = {}
        for idx, item in enumerate(actual):
            result = self.expected.matches(item)
            if not result:
                failures[idx] = result

        if failures:
            return MatchResult.failure(
                "\n".join(
                    ["Non-matching items:"] +
                    ["- at index %d: %s" % (idx, failures[idx].description) for idx in sorted(failures)]
                )
            )
        else:
            return MatchResult.success()


def has_all_items(expected: Any) -> Matcher:
    """Test if all the items of the sequence match expected"""
    return HasAllItems(is_(expected))


class IsIn(Matcher):
    def __init__(self, expected):
        self.expected = expected

    def build_description(self, transformation):
        return transformation("to be in %s" % _jsonify_items(self.expected))

    def matches(self, actual):
        return MatchResult(actual in self.expected, "got %s" % jsonify(actual))


def is_in(expected: Collection) -> Matcher:
    """Test if the sequence contains the expected item"""
    return IsIn(expected)



import sys
from contextlib import contextmanager

from lemoncheesecake.cli.command import Command
from lemoncheesecake.cli.utils import auto_detect_reporting_backends, add_report_path_cli_arg, get_report_path
from lemoncheesecake.reporting import load_report
from lemoncheesecake.reporting.backends.console import print_report_as_test_run
from lemoncheesecake.reporting.console import print_report
from lemoncheesecake.filter import add_result_filter_cli_args, make_result_filter


@contextmanager
def ignore_broken_pipe():
    try:
        yield
    except BrokenPipeError:
        # Broken pipe (example: "lcc report | head"), in that case, simply exit gracefully
        # We just close stderr to avoid Python warnings about ignored exception at the end of the
        # command.
        # See https://stackoverflow.com/questions/26692284/how-to-prevent-brokenpipeerror-when-doing-a-flush-in-python
        # Please note that the ignored exception is related to the stream wrapper of colorama
        sys.stderr.close()


class ReportCommand(Command):
    def get_name(self):
        return "report"

    def get_description(self):
        return "Display a report"

    def add_cli_args(self, cli_parser):
        add_result_filter_cli_args(cli_parser)

        group = cli_parser.add_argument_group("Display report")
        add_report_path_cli_arg(group)
        group.add_argument(
            "--short", "-s", action="store_true", required=False,
            help="Display report as lcc run display test results"
        )
        group.add_argument(
            "--debug", "-d", action="store_true", required=False,
            help="Show debug logs"
        )
        group.add_argument(
            "--explicit", "-e", action="store_true", required=False,
            help="Make all indicators 'explicit' (i.e not only relying on a color-code), "
                 "will be enforced is stdout is redirected"
        )
        group.add_argument(
            "--max-width", "-w", type=int, required=False,
            help="Set a max width for tables (default is current terminal width)"
        )

    def run_cmd(self, cli_args):
        report_path = get_report_path(cli_args)
        report = load_report(report_path, auto_detect_reporting_backends())
        result_filter = make_result_filter(cli_args)

        if cli_args.short:
            print_report_as_test_run(report, result_filter)
        else:
            with ignore_broken_pipe():
                print_report(
                    report, result_filter=result_filter, max_width=cli_args.max_width,
                    show_debug_logs=cli_args.debug,
                    explicit=cli_args.explicit or not sys.stdout.isatty()
                )

        return 0



from collections import defaultdict

from termcolor import colored

from lemoncheesecake.cli.command import Command
from lemoncheesecake.cli.utils import auto_detect_reporting_backends
from lemoncheesecake.reporting.console import test_status_to_color
from lemoncheesecake.reporting import load_report
from lemoncheesecake.filter import add_result_filter_cli_args, make_result_filter
from lemoncheesecake.exceptions import UserError

ORDERED_STATUSES = "failed", "skipped", "disabled", "passed"


class Diff:
    def __init__(self):
        self.added = []
        self.removed = []
        self.status_changed = defaultdict(lambda: defaultdict(list))

    def is_empty(self):
        return not any((self.added, self.removed, self.status_changed))


def compute_diff(report_1_tests, report_2_tests):
    diff = Diff()

    report_2_tests = report_2_tests[:]  # make a copy as a list will be modified
    for report_1_test in report_1_tests:
        # handle removed tests
        try:
            report_2_test = next(test for test in report_2_tests if test.path == report_1_test.path)
        except StopIteration:
            diff.removed.append(report_1_test)
            continue

        # handle status-changed tests
        if report_2_test.status != report_1_test.status:
            diff.status_changed[report_1_test.status][report_2_test.status].append(report_2_test)

        report_2_tests.remove(report_2_test)

    # handle added tests
    diff.added.extend(report_2_tests)

    return diff


def display_diff_type(title, test_entries):
    if not test_entries:
        return

    print("%s (%d):" % (colored(title, attrs=["bold"]), len(test_entries)))
    for test_entry in test_entries:
        print("- %s" % test_entry)
    print()


def render_test_with_status(test):
    return "%s (%s)" % (test.path, colored(test.status, test_status_to_color(test.status)))


def render_test_with_status_changed(test, former_status):
    return "%s (%s => %s)" % (
        test.path,
        colored(former_status, test_status_to_color(former_status)),
        colored(test.status, test_status_to_color(test.status))
    )


def flatten_test_status_changed(status_changed):
    for former_status in ORDERED_STATUSES:
        for new_status in ORDERED_STATUSES:
            for test in status_changed[former_status][new_status]:
                yield test, former_status


def display_diff(diff):
    if diff.is_empty():
        print("There is no difference between the two reports.")
    else:
        display_diff_type(
            "Added tests", list(map(render_test_with_status, diff.added))
        )
        display_diff_type(
            "Removed tests", list(map(render_test_with_status, diff.removed))
        )
        display_diff_type(
            "Status changed",
            [
                render_test_with_status_changed(test, former_status)
                    for test, former_status in flatten_test_status_changed(diff.status_changed)
            ]
        )


class DiffCommand(Command):
    def get_name(self):
        return "diff"

    def get_description(self):
        return "Display differences between two reports"

    def add_cli_args(self, cli_parser):
        add_result_filter_cli_args(cli_parser)
        group = cli_parser.add_argument_group("Diff")
        group.add_argument("report_1_path", help="Report 1 path")
        group.add_argument("report_2_path", help="Report 2 path")

    def run_cmd(self, cli_args):
        reporting_backends = auto_detect_reporting_backends()

        report_1 = load_report(cli_args.report_1_path, reporting_backends)
        report_2 = load_report(cli_args.report_2_path, reporting_backends)
        test_filter = make_result_filter(cli_args)

        report_1_tests = list(filter(test_filter, report_1.all_tests()))
        report_2_tests = list(filter(test_filter, report_2.all_tests()))

        if len(report_1_tests) == 0 and len(report_2_tests) == 0:
            raise UserError("The filter does not match any test on both reports")

        diff = compute_diff(report_1_tests, report_2_tests)
        display_diff(diff)

        return 0



import re
import time

from lemoncheesecake.events import TestSessionSetupEndEvent, TestSessionTeardownEndEvent, \
    TestEndEvent, SuiteSetupEndEvent, SuiteTeardownEndEvent, SuiteEndEvent, SteppedEvent
from lemoncheesecake.reporting.report import ReportLocation

DEFAULT_REPORT_SAVING_STRATEGY = "at_each_failed_test"


def _is_end_of_result_event(event):
    if isinstance(event, TestEndEvent):
        return ReportLocation.in_test(event.test)

    if isinstance(event, SuiteSetupEndEvent):
        return ReportLocation.in_suite_setup(event.suite)

    if isinstance(event, SuiteTeardownEndEvent):
        return ReportLocation.in_suite_teardown(event.suite)

    if isinstance(event, TestSessionSetupEndEvent):
        return ReportLocation.in_test_session_setup()

    if isinstance(event, TestSessionTeardownEndEvent):
        return ReportLocation.in_test_session_teardown()

    return None


def save_at_each_suite_strategy(event, _, __):
    return isinstance(event, SuiteEndEvent)


def save_at_each_test_strategy(event, _, __):
    return _is_end_of_result_event(event) is not None


def save_at_each_failed_test_strategy(event, report, _):
    location = _is_end_of_result_event(event)
    if location:
        result = report.get(location)
        return result and result.status == "failed"
    else:
        return False


def save_at_each_log_strategy(event, _, __):
    return isinstance(event, SteppedEvent)


class SaveAtInterval:
    def __init__(self, interval):
        self.interval = interval

    def __call__(self, event, report, last_saved_time):
        if last_saved_time is None:
            return False

        return last_saved_time + self.interval < time.time()


def make_report_saving_strategy(expression):
    # first, try with a static expression
    static_expressions = {
        "at_end_of_tests": None,  # no need to an intermediate report saving in this case
        "at_each_suite": save_at_each_suite_strategy,
        "at_each_test": save_at_each_test_strategy,
        "at_each_failed_test": save_at_each_failed_test_strategy,
        "at_each_log": save_at_each_log_strategy,
        "at_each_event": save_at_each_log_strategy  # deprecated since 1.4.5, "at_each_log" must be used instead
    }

    try:
        return static_expressions[expression]
    except KeyError:
        pass

    # second, try with "every_Ns"
    m = re.compile(r"^every[_ ](\d+)s$").match(expression)
    if m:
        return SaveAtInterval(int(m.group(1)))

    # ok... nothing we know about
    raise ValueError("Invalid expression '%s' for report saving strategy" % expression)
